/*
 * Copyright (c) 2017 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.fukurou.fileexec;

// import java.io.File;
import java.io.IOException;

import java.nio.file.WatchEvent;
import java.nio.file.Path;
import java.nio.file.PathMatcher;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;							// 7.4.4.0 (2021/06/30)
import java.nio.file.WatchKey;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchService;

import java.util.function.BiConsumer;
// import java.util.concurrent.atomic.AtomicBoolean;		// 7.2.9.4 (2020/11/20) volatile boolean の代替え , // 7.4.4.0 (2021/06/30) 戻す

/**
 * FileWatch は、ファイル監視を行うクラスです。
 *
 *<pre>
 * ファイルが、追加(作成)、変更、削除された場合に、イベントが発生します。
 * このクラスは、Runnable インターフェースを実装しているため、Thread で実行することで、
 * 個々のフォルダの監視を行います。
 *
 *</pre>
 * @og.rev 7.0.0.0 (2017/07/07) 新規作成
 *
 * @version  7.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK1.8,
 */
public class FileWatch implements Runnable {
	private static final XLogger LOGGER= XLogger.getLogger( FileWatch.class.getSimpleName() );		// ログ出力

	/** Path に、WatchService を register するときの作成イベントの簡易指定できるように。 */
	public static final WatchEvent.Kind<Path> CREATE = StandardWatchEventKinds.ENTRY_CREATE ;

	/** Path に、WatchService を register するときの変更イベントの簡易指定できるように。 */
	public static final WatchEvent.Kind<Path> MODIFY = StandardWatchEventKinds.ENTRY_MODIFY ;

	/** Path に、WatchService を register するときの削除イベントの簡易指定できるように。  */
	public static final WatchEvent.Kind<Path> DELETE = StandardWatchEventKinds.ENTRY_DELETE ;

	/** Path に、WatchService を register するときの特定不能時イベントの簡易指定できるように。 */
	public static final WatchEvent.Kind<?>    OVERFLOW = StandardWatchEventKinds.OVERFLOW ;

	// Path に、WatchService を register するときのイベント
	private static final WatchEvent.Kind<?>[] WE_KIND = new WatchEvent.Kind<?>[] {
			CREATE , MODIFY , DELETE , OVERFLOW
	};

	// Path に、WatchService を register するときの登録方法の修飾子(修飾子 なしの場合)
	private static final WatchEvent.Modifier[] WE_MOD_ONE  = new WatchEvent.Modifier[0];	// Modifier なし

	// Path に、WatchService を register するときの登録方法の修飾子(以下の階層も監視対象にします)
	private static final WatchEvent.Modifier[] WE_MOD_TREE = new WatchEvent.Modifier[] {	// ツリー階層
					com.sun.nio.file.ExtendedWatchEventModifier.FILE_TREE
			};

	/** DirWatch でスキャンした場合のｲﾍﾞﾝﾄ名 {@value} */
	public static final String DIR_WATCH_EVENT = "DirWatch";

	/** 7.4.4.0 (2021/06/30) stop() してから、実際に止まるまで 待機する時間 (ms ) */
	public static final int STOP_WATI_TIME = 500 ;

	/** 7.4.4.0 (2021/06/30) stop() してから、実際に止まるまで 待機する回数 */
	public static final int STOP_WATI_CNT = 5 ;

	// 監視対象のフォルダ
	private final Path dirPath ;

	// 監視方法
	private final boolean	useTree ;
	private final WatchEvent.Modifier[] extModifiers ;

	// callbackするための、関数型インターフェース(メソッド参照)
	private BiConsumer<String,Path> action = (event,path) -> System.out.println( "Event=" + event + " , Path=" + path ) ;

	// Path に、WatchService を register するときのイベント
	private WatchEvent.Kind<?>[] weKind = WE_KIND ;						// 初期値は、すべて

	// パスの照合操作を行うPathMatcher の初期値
	private final PathMatcherSet pathMchSet = new PathMatcherSet();		// PathMatcher インターフェースを継承

	// DirWatchのパスの照合操作を行うPathMatcher の初期値
	private final PathMatcherSet dirWatchMch = new PathMatcherSet();	// PathMatcher インターフェースを継承

	// 何らかの原因でイベントもれした場合、フォルダスキャンを行います。
	private boolean		useDirWatch	= true;								// 初期値は、イベント漏れ監視を行います。
	private DirWatch	dWatch ;										// DirWatch のstop時に呼び出すための変数
	private Thread		thread ;										// 停止するときに呼び出すため

	private volatile boolean running ;									// 状態とThreadの停止に使用する。 // 7.4.4.0 (2021/06/30) 復活
//	private final AtomicBoolean running = new AtomicBoolean();			// 7.2.9.4 (2020/11/20) volatile boolean の代替え ( 状態とThreadの停止に使用する。)

	/**
	 * 処理対象のフォルダのパスオブジェクトを指定して、ファイル監視インスタンスを作成します。
	 *
	 * ここでは、指定のフォルダの内のファイルのみ監視します。
	 * これは、new FileWatch( dir , false ) とまったく同じです。
	 *
	 * @param dir	処理対象のフォルダオブジェクト
	 */
	public FileWatch( final Path dir ) {
		this( dir , false );
	}

	/**
	 * 処理対象のフォルダのパスオブジェクトと、監視対象方法を指定して、ファイル監視インスタンスを作成します。
	 *
	 * useTree を true に設定すると、指定のフォルダの内のフォルダ階層を、すべて監視対象とします。
	 *
	 * @param dir	処理対象のフォルダのパスオブジェクト
	 * @param useTree	フォルダツリーの階層をさかのぼって監視するかどうか(true:フォルダ階層を下る)
	 */
	public FileWatch( final Path dir , final boolean useTree ) {
		dirPath		 = dir ;
		this.useTree = useTree;
		extModifiers = useTree ? WE_MOD_TREE : WE_MOD_ONE ;
	}

	/**
	 * 指定のイベントの種類のみ、監視対象に設定します。
	 *
	 * ここで指定したイベントのみ、監視対象になり、callback されます。
	 * 第一引数は、イベントの種類(ENTRY_CREATE,ENTRY_MODIFY,ENTRY_DELETE,OVERFLOW)
	 *
	 * @param	kind 監視対象に設定するイベントの種類
	 * @see		java.nio.file.StandardWatchEventKinds
	 */
	public void setEventKinds( final WatchEvent.Kind<?>... kind ) {
		if( kind != null && kind.length > 0 ) {
			weKind = kind;
		}
	}

	/**
	 * 指定のパスの照合操作で、パターンに一致したパスのみ、callback されます。
	 *
	 * ここで指定したパターンの一致を判定し、一致した場合は、callback されます。
	 * 指定しない場合は、すべて許可されたことになります。
	 * なお、#setPathEndsWith(String...) と、この設定は同時には行うことは出来ません。
	 * (最後に登録した条件が、適用されます。)
	 *
	 * @param	pathMch パスの照合操作のパターン
	 * @see		java.nio.file.PathMatcher
	 * @see		#setPathEndsWith(String...)
	 */
	public void setPathMatcher( final PathMatcher pathMch ) {
		pathMchSet.addPathMatcher( pathMch );
	}

	/**
	 * 指定のパスが、指定の文字列と、終端一致(endsWith) したパスのみ、callback されます。
	 *
	 * これは、#setPathMatcher(PathMatcher) の簡易指定版です。
	 * 指定の終端文字列(一般には拡張子)のうち、ひとつでも一致すれば、true となりcallback されます。
	 * 指定しない場合(null)は、すべて許可されたことになります。
	 * 終端文字列の判定には、大文字小文字の区別を行いません。
	 * なお、#setPathMatcher(PathMatcher) と、この設定は同時には行うことは出来ません。
	 * (最後に登録した条件が、適用されます。)
	 *
	 * @param	endKey パスの終端一致のパターン
	 * @see		#setPathMatcher(PathMatcher)
	 */
	public void setPathEndsWith( final String... endKey ) {
		pathMchSet.addEndsWith( endKey );
	}

	/**
	 * イベントの種類と、ファイルパスを、引数に取る BiConsumer ダオブジェクトを設定します。
	 *
	 * これは、関数型インタフェースなので、ラムダ式またはメソッド参照の代入先として使用できます。
	 * イベントが発生したときの イベントの種類と、そのファイルパスを引数に、accept(String,Path) メソッドが呼ばれます。
	 * 第一引数は、イベントの種類(ENTRY_CREATE,ENTRY_MODIFY,ENTRY_DELETE,OVERFLOW)
	 * 第二引数は、ファイルパス(監視フォルダで、resolveされた、正式なフルパス)
	 *
	 * @param	act 2つの入力(イベントの種類 とファイルパス) を受け取る関数型インタフェース
	 * @see		BiConsumer#accept(Object,Object)
	 */
	public void callback( final BiConsumer<String,Path> act ) {
		if( act != null ) {
			action = act ;
		}
	}

	/**
	 * 何らかの原因でイベントを掴み損ねた場合に、フォルダスキャンするかどうかを指定します。
	 *
	 * スキャン開始の遅延時間と、スキャン間隔、ファイルのタイムスタンプとの比較時間等は、
	 * DirWatch の初期値をそのまま使用するため、ここでは指定できません。
	 * 個別に指定したい場合は、このﾌﾗｸﾞをfalse にセットして、個別に、DirWatch を作成してください。
	 * このメソッドでは、#setPathEndsWith( String... )や、#setPathMatcher( PathMatcher ) で
	 * 指定した条件が、そのまま適用されます。
	 *
	 * @param	flag フォルダスキャンするかどうか(true:する/false:しない)
	 * @see		DirWatch
	 */
	public void setUseDirWatch( final boolean flag ) {
		useDirWatch = flag;
	}

	/**
	 * 何らかの原因でイベントを掴み損ねた場合の、フォルダスキャンの対象ファイルの拡張子を指定します。
	 *
	 * このメソッドを使用する場合は、useDirWatch は、true にセットされます。
	 * スキャン開始の遅延時間と、スキャン間隔、ファイルのタイムスタンプとの比較時間等は、
	 * DirWatch の初期値をそのまま使用するため、ここでは指定できません。
	 * このメソッドでは、DirWatch 対象の終端パターンを独自に指定できますが、FileWatch で
	 * で指定した条件も、クリアされるので、含める必要があります。
	 *
	 * @param	endKey パスの終端一致のパターン
	 * @see		DirWatch
	 */
	public void setDirWatchEndsWith( final String... endKey ) {
		if( endKey != null && endKey.length > 0 ) {
			useDirWatch = true;					// 対象があれば、実行するが、true になる。

			dirWatchMch.addEndsWith( endKey );
		}
	}

	/**
	 * このファイル監視で、最後に処理した結果が、エラーの場合に、true を返します。
	 *
	 * 通常は、対象フォルダが見つからない場合や、フォルダスキャン（DirWatch）で
	 * エラーが発生した場合に、true にセットされます。
	 * また、stop() メソッドが呼ばれた場合も、true にセットされます。
	 *
	 * @og.rev 7.2.9.4 (2020/11/20) PMD:volatile boolean の代替え。
	 *
	 * @return	エラー状態(true:エラー,false:正常)
	 */
	public boolean isErrorStatus() {
		// DirWatch を使用している場合は、その結果も加味します。
//		return isError || dWatch != null && dWatch.isErrorStatus() ;
		return !running || dWatch != null && dWatch.isErrorStatus() ;			// 7.4.4.0 (2021/06/30) 復活

//		return !running.get() || dWatch != null && dWatch.isErrorStatus() ;		// 7.2.9.4 (2020/11/20)
	}

	/**
	 * フォルダの監視を開始します。
	 *
	 * @og.rev 7.2.9.4 (2020/11/20) PMD:volatile boolean の代替え。
	 * @og.rev 7.4.4.0 (2021/06/30) stop() してから、実際に止まるまでの時間、待機します。
	 *
	 * 自身を、Threadに登録して、Thread#start() を実行します。
	 * 内部の Thread オブジェクトがなければ、新しく作成します。
	 * すでに、実行中の場合は、何もしません。
	 * 条件を変えて、実行したい場合は、stop() メソッドで、一旦スレッドを
	 * 停止させてから、再び、#start() メソッドを呼び出してください。
	 */
	public void start() {
		// 7.4.4.0 (2021/06/30) stop() してから、実際に止まるまでの時間、待機します。
		int cnt = 0;
		while( running ) {
			cnt++ ;
			try{ Thread.sleep( STOP_WATI_TIME ); } catch( final InterruptedException ex ){}
			if( cnt >= STOP_WATI_CNT ) {	// ループ後も、まだ、stop() 出来ていない場合。
				LOGGER.warning( () -> "FileWatch Stop Error : [" + dirPath + "]" );
			}
		}

		running = true;				// 7.4.4.0 (2021/06/30) 復活

		if( thread == null ) {
			thread = new Thread( this );
//			running = true;
//			running.set( true );	// 7.2.9.4 (2020/11/20)
			thread.start();			// running=true; を先に行わないと、すぐに終了してしまう。
		}

		// 監視漏れのファイルを、一定時間でスキャンする
		if( useDirWatch ) {
			dWatch = new DirWatch( dirPath,useTree );
			if( dirWatchMch.isEmpty() ) {			// 初期値は、未登録時は、本体と同じPathMatcher を使用します。
				dWatch.setPathMatcher( pathMchSet );
			}
			else {
				dWatch.setPathMatcher( dirWatchMch );
			}
			dWatch.callback( path -> action.accept( DIR_WATCH_EVENT , path ) ) ;	// BiConsumer<String,Path> を Consumer<Path> に変換しています。
			dWatch.start();
		}
	}

	/**
	 * フォルダの監視を終了します。
	 *
	 * 自身を登録しているThreadに、割り込みをかけるため、
	 * Thread#interrupt() を実行します。
	 * フォルダ監視は、ファイル変更イベントが発生するまで待機していますが、
	 * interrupt() を実行すると、強制的に中断できます。
	 * 内部の Thread オブジェクトは、破棄するため、再び、start() メソッドで
	 * 実行再開することが可能です。
	 *
	 * @og.rev 7.2.9.4 (2020/11/20) PMD:volatile boolean の代替え。
	 * @og.rev 7.4.4.0 (2021/06/30) thread の存在有無にかかわらず、running は停止状態にする。
	 */
	public void stop() {
		// 7.4.4.0 (2021/06/30) thread の存在有無にかかわらず、running は停止状態にする。
		running = false;		// 7.4.4.0 (2021/06/30) 復活

		if( thread != null ) {
//			running = false;
	//		running.set( false );		// 7.2.9.4 (2020/11/20)
			thread.interrupt();
	//		thread = null;			1.1.0 (2018/02/01) stop() 時に null を入れると、interrupt() 後の処理が継続できなくなる。
	//		なので、run()の最後に、thread = null を入れておきます。
		}

		if( dWatch != null ) {
			dWatch.stop();
			dWatch = null;
		}
	}

	/**
	 * Runnableインターフェースのrunメソッドです。
	 *
	 * 規定のスケジュール時刻が来ると、呼ばれる runメソッドです。
	 *
	 * @og.rev 7.2.5.0 (2020/06/01) LOGGERを使用します。
	 * @og.rev 7.2.9.4 (2020/11/20) PMD:volatile boolean の代替え。
	 */
	@Override
	public void run() {
		try {
			execute();
		}
		catch( final IOException ex ) {
			// MSG0102 = ﾌｧｲﾙ監視に失敗しました。 Path=[{0}]
//			MsgUtil.errPrintln( ex , "MSG0102" , dirPath );
			final String errMsg = "FileWatch#run : Path=" + dirPath ;
			LOGGER.warning( ex , "MSG0102" , errMsg );
		}
		catch( final Throwable th ) {
			// MSG0021 = 予期せぬｴﾗｰが発生しました。\n\tﾒｯｾｰｼﾞ=[{0}]
//			MsgUtil.errPrintln( th , "MSG0021" , toString() );
			final String errMsg = "FileWatch#run : Path=" + dirPath ;
			LOGGER.warning( th , "MSG0021" , errMsg );
		}
		finally {
			running = false;			// 7.4.4.0 (2021/06/30) 停止条件だが、予期せぬエラーで停止した場合も、設定する。
			thread  = null;				// 7.2.5.0 (2020/06/01) 停止処理
//			running = false;
//			running.set( false );		// 7.2.9.4 (2020/11/20)
		}
	}

	/**
	 * runメソッドから呼ばれる、実際の処理。
	 *
	 * try ･･･ catch( Throwable ) 構文を、runメソッドの標準的な作りにしておきたいため、
	 * あえて、実行メソッドを分けているだけです。
	 *
	 * @og.rev 6.8.1.5 (2017/09/08) LOGGER.debug 情報の追加
	 * @og.rev 7.2.9.4 (2020/11/20) PMD:volatile boolean の代替え
	 * @og.rev 8.0.0.0 (2021/07/01) dirPathのsynchronized作成
	 */
	private void execute() throws IOException {
		// ファイル監視などの機能は新しいNIO2クラスで拡張されたので
		// 旧File型から、新しいPath型に変換する.
		LOGGER.info( () -> "FileWatch Start: " + dirPath );

		// デフォルトのファイル・システムを閉じることはできません。(UnsupportedOperationException がスローされる)
		// なので、try-with-resources 文 (AutoCloseable) に、入れません。
//		final FileSystem fs = dirPath.getFileSystem();			// フォルダが属するファイルシステムを得る()
		final FileSystem fs = FileSystems.getDefault();			// 7.4.4.0 (2021/06/30) 上記と同じｵﾌﾞｼﾞｪｸﾄだから。

		// try-with-resources 文 (AutoCloseable)
		// ファイルシステムに対応する監視サービスを構築する.
		// (一つのサービスで複数の監視が可能)
		try( WatchService watcher = fs.newWatchService() ) {
			// フォルダに対して監視サービスを登録する.
			final WatchKey watchKey = dirPath.register( watcher , weKind , extModifiers );

			// 監視が有効であるかぎり、ループする.
			// (監視がcancelされるか、監視サービスが停止した場合はfalseとなる)
			try{
				boolean flag = true;
				while( flag && running ) {										// 7.4.4.0 (2021/06/30) 復活
//				while( flag && running.get() ) {								// 7.2.9.4 (2020/11/20)
					// スレッドの割り込み = 終了要求を判定する.
			//		if( Thread.currentThread().isInterrupted() ) {
			//			throw new InterruptedException();
			//		}

					// take は、ファイル変更イベントが発生するまで待機する.
					final WatchKey detectKey = watcher.take();			// poll() は、キューが空の場合はブロックせずに null を返す

					// イベント発生元を判定する
//					if( detectKey.equals( watchKey ) ) {
					if( watchKey.equals( detectKey ) ) {				// 8.0.0.0 (2021/07/01) 入れ替え(null対応)
						// 発生したイベント内容をプリントする.
						for( final WatchEvent<?> event : detectKey.pollEvents() ) {
							// 追加・変更・削除対象のファイルを取得する.
							// (ただし、overflow時などはnullとなることに注意)
							final Path path = (Path)event.context();
							if( path != null && pathMchSet.matches( path ) ) {
								final Path fpath = dirPath.resolve( path );
								synchronized( dirPath ) {				// 8.0.0.0 (2021/07/01) dirPathのsynchronized作成
									if( dWatch == null || dWatch.setAdd( fpath) ) {		// このセット内に、指定された要素がなかった場合はtrue
										action.accept( event.kind().name() , fpath );
									}
									else {
										// CREATE と MODIFY などのイベントが連続して発生するケースへの対応
										LOGGER.info( () -> "WatchEvent Duplication: " + fpath );
									}
								}
							}
						}
					}

					// イベントの受付を再開する.
					if( detectKey != null ) {		// 8.0.0.0 (2021/07/01) null対応
						detectKey.reset();
					}

					if( dWatch != null ) {
						dWatch.setClear();			// Path重複ﾁｪｯｸ用のSetは、一連のｲﾍﾞﾝﾄ完了時にｸﾘｱしておきます。
					}

					// 監視サービスが活きている、または、スレッドの割り込み( = 終了要求)がないことを、をチェックする。
					flag = watchKey.isValid() && !Thread.currentThread().isInterrupted() ;

					// 7.4.4.0 (2021/06/30) ※ 63ﾌｫﾙﾀﾞ以上は、監視できない？(Tomcat上では？)
					if( !watchKey.isValid() ) {
						LOGGER.warning( () -> "FileWatch No isValid : [" + dirPath + "]" );
					}
				}
			}
			catch( final InterruptedException ex ) {
//				LOGGER.warning( () -> "【WARNING】 FileWatch Canceled:" + dirPath );
				LOGGER.warning( () -> "FileWatch Canceled : [" + dirPath + "]" );
			}
			finally {
				// スレッドの割り込み = 終了要求なので監視をキャンセルしループを終了する。
				if( watchKey != null ) {
					watchKey.cancel();
				}
			}
		}
		// FileSystemの実装(sun.nio.fs.WindowsFileSystem)は、close() 未サポート
		catch( final UnsupportedOperationException ex ) {
			LOGGER.warning( () -> "FileSystem close : [" + dirPath + "]" );
		}

		// 7.4.4.0 (2021/06/30) 念のため、入れておきます。
		catch( final Throwable th ) {
			// MSG0021 = 予期せぬｴﾗｰが発生しました。\n\tﾒｯｾｰｼﾞ=[{0}]
			final String errMsg = "FileWatch#execute : Path=" + dirPath ;
			LOGGER.warning( th , "MSG0021" , errMsg );
		}

//		LOGGER.info( () -> "FileWatch End: " + dirPath );
		LOGGER.info( () -> "FileWatch End : [" + dirPath + "]" );

//		thread  = null;					// 1.1.0 (2018/02/01) 停止処理
	//	isError = true;					// 何らかの原因で停止すれば、エラーと判断します。
	}

	/**
	 *このオブジェクトの文字列表現を返します。
	 *
	 * @return	このオブジェクトの文字列表現
	 */
	@Override
	public String toString() {
		return getClass().getSimpleName() + ":" + dirPath + " , " + DIR_WATCH_EVENT + "=[" + useDirWatch + "]" ;
	}
}
